#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define TOK_MAX		20

#define TOK_BLANK	0
#define TOK_LABEL	1
#define TOK_ARITH	2
#define TOK_PAREN	3


#define FREE(x) if (x) { free(x); x = NULL; }


int tokIsLabel(const char *ch)
{
  if (isalnum(*ch)) return 1;
  if (*ch == '@') return 1;
  if (*ch == '_') return 1;
  if (*ch == '$') return 1;
  return 0;
}


int tokIsArith(const char *ch)
{
  if (*ch == '+') return 1;
  if (*ch == '-') return 1;
  if (*ch == '*') return 1;
  if (*ch == '/') return 1;
  if (*ch == '%') return 1;
  if (*ch == '<') return 1;
  if (*ch == '>') return 1;
  if (*ch == '&') return 1;
  if (*ch == '|') return 1;
  if (*ch == '^') return 1;
  return 0;
}


int tokIsParen(const char *ch)
{
  if (*ch == '(') return 1;
  if (*ch == ')') return 1;
  return 0;
}


int tokType(const char *ch)
{
  if (isblank(*ch)) return TOK_BLANK;
  if (tokIsLabel(ch)) return TOK_LABEL;
  if (tokIsArith(ch)) return TOK_ARITH;
  if (tokIsParen(ch)) return TOK_PAREN;
  return TOK_BLANK;	// Unknown characters treated as blanks
}


char **tokSplit(const char *str)
{
  int argc = 0;
  char **argv;
  const char *s;
  const char *t;
  int i;
  int toktype = TOK_BLANK;
  int newtt;

  argv = (char **)calloc(TOK_MAX, sizeof(char *));

  t = s = str;
  i = 1;
  while (*s != '\0') {
    newtt = tokType(s);
    // fprintf(stderr, "(%d) %d %s\n", toktype, newtt, s);
    if (newtt != toktype) {
      // Blanks seperate token but are not tokens themselves.
      if (toktype != TOK_BLANK) {
        argv[argc] = strndup(t, i-1);
        argc++;
      }
      t = s;
      i = 1;
      toktype = newtt;
      if (argc >= (TOK_MAX - 1)) {
        fprintf(stderr, "WARNING: Max tokens exceeded. Skipping remainder.\n");
        break;
      }
    }  // if new token
    s++;
    i++;
  }  // while not end of string
  if ((toktype != TOK_BLANK) && (argc < TOK_MAX)) argv[argc++] = strndup(t, i);
  argv[argc] = NULL;
  return argv;
}


void tokFree(char **tokv)
{
  int i;

  i=0;
  while (tokv[i]) {
    FREE(tokv[i]);
    i++;
  }
  free(tokv);
}


int tokCount(char ** tokv)
{
  int count = 0;
  while (tokv[count] != NULL) count++;
  return count;
}


// Just to wrap my head around this...
//   tokv is a pointer to an array of character pointers (strings).
//   What I need: passed tokv itself should not change, but
//       the list of character pointers it points to should change
//       and those changes should be returned.
void tokShiftL(char ** tokv, int pos, int count)
{
  int j;

  j = 0;
  do {
    // fprintf(stderr, "    tokShiftL: tokv[%d] %s", pos+j, tokv[pos+j]);
    if (j<count) FREE(tokv[pos+j]);
    tokv[pos+j] = tokv[pos+j+count];
    // fprintf(stderr, " <-- %s\n", tokv[pos+j]);
    if (tokv[pos+j] == NULL) break;
    j++;
  } while (1);
}


char **tokDup(char **tokv, int start, int end)
{
  int len;
  int i;
  char **argv;

  len = (end - start) + 2;
  argv = (char **)calloc(len, sizeof(char *));
  for (i=start; i<=end; i++) {
    argv[i-start] = strdup(tokv[i]);
  }
  argv[i-start] = NULL;
  return argv;
}


void tokDump(char **tokv)
{
  int i = 0;
  while (tokv[i]) {
    printf("%d: '%s'\n", i, tokv[i]);
    i++;
  }
}

